from typing import Optional, Callable, Union, Any
from dramatiq import Actor
from celery.app.task import Task

DecoratorInner = Union[Task, Actor, Callable[[], None]]

def cron(
    crontab: str, job_id: Optional[str] = None
) -> Callable[[DecoratorInner], DecoratorInner]: ...



"""
dispatcher.py

The dispatcher loop that handles processing scheduled events and dispatching
tasks either to a task framework (e.g. Dramatiq or Celery) or to a pool of
sub-processes.
"""

#  Copyright (c) 2020 Daniel Sloan, Lucid Horizons Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#

import logging
import signal

from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.executors.pool import ProcessPoolExecutor, ThreadPoolExecutor
from pytz import utc

from django_cadence.apps import JOBS

logger = logging.getLogger(__name__)


def start_dispatcher(processes):
    """Start the dispatcher loop"""

    logger.info("Starting job dispatcher loop")
    # FIXME: Get timezone from django ?
    tz = utc
    jobstores = {
        "default": MemoryJobStore(),
        # FIXME: Add support for using Django model as job store in future
        # 'django': DjangoJobStore()
    }
    if processes > 1:
        # Process pool; good for cases where we are not using Celery or Dramatiq and need to start a process for each
        #  task to be handled in this process.
        executors = {
            "default": ProcessPoolExecutor(max_workers=processes),
        }
    else:
        # Thread pool; for very small thread-safe environments or where all we do is send tasks to Celery or Dramatiq
        executors = {"default": ThreadPoolExecutor(max_workers=8)}

    scheduler = BlockingScheduler(
        logger=logger, timezone=tz, executors=executors, jobstores=jobstores
    )

    for trigger, func, job_id in JOBS:
        logger.debug("Adding task: %s - schedule %r", job_id, trigger)
        scheduler.add_job(
            func,
            trigger=trigger,
            name=job_id,
            id=job_id,
            max_instances=1,
            replace_existing=True,
        )

    def shutdown(signum, frame):
        scheduler.shutdown()

    signal.signal(signal.SIGINT, shutdown)
    signal.signal(signal.SIGTERM, shutdown)

    logger.debug("Starting blocking main loop for cron server...")
    scheduler.start()
    logger.debug("Shutting down main loop")



#  Copyright (c) 2020 Daniel Sloan, Lucid Horizons Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#

from .scheduling import cron

# Allows "legacy" style use of "django_cadence" as the app, instead of requiring the full AppConfig
default_app_config = "django_cadence.apps.DjangoCadenceConfig"

__all__ = ["default_app_config", "cron"]



"""
apps.py

The AppConfig for the django-cadence app, which provides initialization
of the known job list.
"""

#  Copyright (c) 2020 Daniel Sloan, Lucid Horizons Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#

import logging
from importlib import import_module

from django.apps import AppConfig
from django.apps import apps

logger = logging.getLogger(__name__)
JOBS = []


class DjangoCadenceConfig(AppConfig):
    name = "django_cadence"
    loaded = False

    def ready(self):
        if not self.loaded:
            self.loaded = True

            # For each app in INSTALLED_APPS, make sure we import the tasks module, so that any cron tasks are found.
            for app_config in apps.get_app_configs():
                try:
                    mod = import_module(f"{app_config.name}.tasks")
                    logger.info("Loaded module %s.tasks: %s", app_config.name, mod)
                except ModuleNotFoundError:
                    logger.debug("No tasks module in %s", app_config.name)
                except ImportError:
                    logger.error("Error loading {app_config.name}.tasks", exc_info=True)



"""
scheduling.py

Decorators to wrap django functions and make them scheduled.

"""

# Copyright (c) 2020 Daniel Sloan, Lucid Horizons Pty Ltd

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Callable, Union, Optional, TYPE_CHECKING

from apscheduler.triggers.cron import CronTrigger

from .apps import JOBS


def cron(crontab, job_id=None):
    """
    Wrap a regular function, Celery task, or Dramatiq actor in a cron schedule.
    """
    trigger = CronTrigger.from_crontab(crontab)

    def decorator(f):
        """Create a scheduled trigger for the decorated callable"""

        # We might be getting called as a wrapper of a regular callable (function, class, etc)
        # Or we might be getting called as a wrapper of a Dramatiq or Celery task.
        # We'll start out just adding the callable to the JOBS list, but for example Dramatiq
        # actors should be invoked as "<f>.send()", and Celery tasks should be invoked as
        # "<f>.apply_async()".

        if hasattr(f, "send"):
            # It's (probably) a Dramatiq Actor
            func_call = f.send
            job_name = job_id or f"{f.fn.__module__}.{f.fn.__name__}"
        elif hasattr(f, "apply_async"):
            # It's a Celery task or shared_task
            func_call = f.apply_async
            job_name = job_id or f"{f.name}"
        else:
            # It's something else, probably just a regular callable - call it directly and use repr() as the name
            func_call = f
            job_name = job_id or f"{f.__module__}.{f.__name__}"
        JOBS.append((trigger, func_call, job_name))
        return f

    return decorator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_cadence',
 'django_cadence.management',
 'django_cadence.management.commands']

package_data = \
{'': ['*']}

install_requires = \
['apscheduler>=3.6.0,<4.0.0', 'django>=2.2.0,<3.2']

extras_require = \
{'celery': ['celery>=4.4.0,<5.0.0'],
 'dramatiq': ['django_dramatiq>=0.8.0,<1.0.0']}

setup_kwargs = {
    'name': 'django-cadence',
    'version': '0.1.2',
    'description': 'A package to provide periodic scheduled task support in Django',
    'long_description': '# django_cadence\n\nA reusable Django app that provides scheduling of asynchronous tasks (optionally as asynchronous tasks via Dramatiq or Celery)\n\n![Latest GitHub Release](https://img.shields.io/github/v/release/LucidDan/django-cadence?sort=semver&style=plastic)\n![Latest Pypi Release](https://img.shields.io/pypi/v/django-cadence.svg?style=plastic)\n![License](https://img.shields.io/github/license/LucidDan/django-cadence?style=plastic)\n![Build](https://img.shields.io/travis/:LucidDan/django-cadence?style=plastic)\n\n![Photo of a calendar on a wall](https://unsplash.com/photos/PypjzKTUqLo "Photo by Roman Bozhko")\n\n## Introduction\n\nThis is a simple package, so far at least, that provides some scheduling capabilities in Django.\nIt can do this standalone, or in cooperation with [Dramatiq] or [Celery].\n\n\n## Links and References\n\n* [Releases - PyPI](https://pypi.org/django-cadence)\n* [Releases - GitHub](https://github.com/LucidDan/django-cadence/releases)\n* [GitHub Issues]\n* [Source](https://github.com/LucidDan/django-cadence/)\n* [ChangeLog](CHANGELOG.md)\n* [License](LICENSE.md)\n* [Contributing](CONTRIBUTING.md)\n\n## What\'s New\n\nCurrent version: 0.1.2\n\nTesting travis deployment, and updating documentation.\n\nSee the [changelog](CHANGELOG.md) for more information on recent changes.\n\n## Supported Releases\n\nSee [SECURITY.md] for releases which are currently supported for security updates.\n\nIn general, the current major release is supported for all bug and security fixes, and the previous major release is supported for security fixes only.\n\n## Dependencies\n\nRequired:\n * [Python]() v3.6.x - v3.8.x\n * [Django]() v2.2.x - v3.0.x\n * [APScheduler]() v3.6.x\n \nRecommended:\n * [Dramatiq]() v1.8.x - v1.9.x\n  (or)\n * [Celery]() v4.4.x\n\n\n## Installing\n\nUse pip or another pip-based tool in combination with PyPI to install this package.\n\n\n## Getting Support\n\nFirst, unless it is extremely urgent, please check the FAQ, Known Issues section, and general documentation before looking for help.\n\nThere are a few avenues available for support. In general, if you\'re really confident you\'ve found a bug, log an issue directly via GitHub.\n\n* [Log an issue via GitHub][GitHub Issues]\n\n\n## FAQ\n\nThis section contains some frequent questions and answers, be aware that if you did not read this section and ask one of these questions on the forum or chat server, you might get redirected back to read this with varying levels of politeness or irritation. ;-)\n\n### When will the mod be updated for Django version X, Python version Y, etc?\n\nWhen it is done. Pull requests are welcome.\n\n\n## Known Issues\n\nThis section attempts to sum up known issues in a easier-to-consume way. For the most up-to-date information on this, it might be worth checking [GitHub Issues]\n\n* There are currently no known issues.\n\n\n## Contributing\n\nSee [CONTRIBUTING.md] for information on how you can help contribute to this project.\n\n\n## Special Thanks\n\nHere is a non-exhaustive list of sources, people, etc that deserve some thanks for bringing this project into existence:\n\n* []()\n\n\n\n[GitHub Issues]: https://github.com/LucidDan/django-cadence/issues\n[Dramatiq]: https://dramatiq.io/\n[Celery]: https://docs.celeryproject.org/\n\n',
    'author': 'Dan Sloan',
    'author_email': 'dan@luciddan.com',
    'maintainer': 'Dan Sloan',
    'maintainer_email': 'dan@luciddan.com',
    'url': 'https://luciddan.github.io/django-cadence/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
runcadence.py

Management command to run the main scheduler loop. This runs in a separate process
and is responsible for making sure scheduled tasks are executed at the right time.

Typically, it is running in an environment with Dramatiq or Celery, and mostly just
 dispatches tasks to the worker processes. However it might also run standalone,
 in which case it will dispatch the tasks to its own pool of multiprocessing workers.
"""

#  Copyright (c) 2020 Daniel Sloan, Lucid Horizons Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#

import multiprocessing

from django.core.management.base import BaseCommand
from django_cadence.dispatcher import start_dispatcher

CPU_COUNT = multiprocessing.cpu_count()


class Command(BaseCommand):
    help = "Runs Cadence scheduled tasks."

    def add_arguments(self, parser):
        parser.add_argument(
            "--processes",
            "-p",
            default=CPU_COUNT,
            type=int,
            help="The number of processes to run (default: %d)." % CPU_COUNT,
        )

    def handle(self, processes, **options):
        """Handle a call to the command via manage.py"""

        # Set up and run the job dispatcher
        start_dispatcher(processes=processes)
